"""
Reinfrocement Learning Algorithms
"""

from rl_replicas.algorithms.ddpg import DDPG
from rl_replicas.algorithms.ppo import PPO
from rl_replicas.algorithms.td3 import TD3
from rl_replicas.algorithms.trpo import TRPO
from rl_replicas.algorithms.vpg import VPG
