"""
Policies
"""
from rl_replicas.policies.categorical_policy import CategoricalPolicy
from rl_replicas.policies.deterministic_policy import DeterministicPolicy
from rl_replicas.policies.gaussian_policy import GaussianPolicy
from rl_replicas.policies.policy import Policy
from rl_replicas.policies.random_policy import RandomPolicy
from rl_replicas.policies.stochastic_policy import StochasticPolicy
